
#include <propeller2.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>

#define _DIAG

enum {
//	_xinfreq = 20_000_000,
	_xtlfreq = 20_000_000,
	_clkfreq = 4_000_000,

	DOWNLOAD_BAUD = 230_400,
	DEBUG_BAUD = DOWNLOAD_BAUD,

	SD_BASE_PIN = 16,
	CLK_DIV = 20,

	PIN_DAT    = SD_BASE_PIN | 3<<6,
	PIN_DAT0   = SD_BASE_PIN+0,
	PIN_DAT1   = SD_BASE_PIN+1,
	PIN_DAT2   = SD_BASE_PIN+2,
	PIN_DAT3   = SD_BASE_PIN+3,
	PIN_CMD    = SD_BASE_PIN+4,
	PIN_CLK    = SD_BASE_PIN+5,
	PIN_LED  = SD_BASE_PIN+6,  // command input smartpin, also outputs to Activity LED
	PIN_CD     = SD_BASE_PIN+7,  // Card-Detect switch input, doubles up as PIN_CMD Start-Bit detect
};


static uint32_t  rca;
static uint8_t  resp[16];
static uint8_t  buff[2000];

/*
    SD Responses
   ==============
    R1   48 bits    Status (32-bit)
    R1b  48 bits, and busy on DAT0
    R2  136 bits    CID or CSD
    R3   48 bits    OCR (32-bit)
    R6   48 bits    RCA (16-bit) and Status (16-bit)
    R7   48 bits    volts (4-bit)
*/

#ifdef _DIAG
static uint32_t  crc7( uint8_t * buf, size_t len )    // SD spec 4.5
{
    uint32_t  val, crc = 0;

    __asm {
// Reference code courtesy of Ariba
cr7lp
		rdword	val, buf
		add	buf, #2
		movbyts	val, #0b00_01_10_11
		setq	val
		crcnib	crc, #0x48    // CCITT polynomial is 1 + x3 + x7 (0x09 reversed for CRCNIB)
		crcnib	crc, #0x48
		sub	len, #1   wcz
	if_ne	crcnib	crc, #0x48
	if_ne	crcnib	crc, #0x48
	if_ne	djnz	len, #cr7lp

		rev	crc         // correct the bit order to match standard
		shr	crc, #24
		or	crc, #1
    }
    return crc;
}
#endif


static uint32_t rxblock( uint8_t * buf )
{
    uint32_t  i, j, shiftl, shiftr;

    _pinf( SD_BASE_PIN | 7<<6 );
    _wrpin( PIN_DAT2, P_SYNC_RX |    // setup smartpin for DAT0 pin
                    (PIN_DAT0 - PIN_DAT2 & 7)<<28 |
                    (PIN_CLK - PIN_DAT2 & 7)<<24 );
    _wrpin( PIN_DAT3, P_SYNC_RX |    // setup smartpin for DAT1 pin
                    (PIN_DAT1 - PIN_DAT3 & 7)<<28 |
                    (PIN_CLK - PIN_DAT3 & 7)<<24 );
    _wrpin( PIN_LED, P_SYNC_RX |    // setup smartpin for DAT3 pin
                    (PIN_DAT3 - PIN_LED & 7)<<28 |
                    (PIN_CLK - PIN_LED & 7)<<24 );
    _wxpin( PIN_DAT2, 31 );    // DAT0 32-bit shifter/buffer, pre-rising clock sampling
    _wxpin( PIN_DAT3, 31 );    // DAT1 32-bit shifter/buffer, pre-rising clock sampling
    _wxpin( PIN_LED, 31 );    // DAT3 32-bit shifter/buffer, pre-rising clock sampling

    _wrpin( PIN_CMD, P_SYNC_RX | (PIN_CLK - PIN_CMD & 7)<<24 );    // setup command responce smartpin mode
    _wxpin( PIN_CMD, 15 );    // CMD 16-bit shifter/buffer, pre-rising clock sampling

    __asm {
		dirh	#PIN_CLK
		wypin	#100, #PIN_CLK
		dirh	#PIN_CMD
		setse1	#001<<6 | PIN_CMD
.loop1
		waitse1
		rdpin	pb, #PIN_CMD
		rev	pb
		signx	pb, #15
		not	shiftr, pb   wz
	if_z	jmp	#.loop1

		encod	shiftr
		add	shiftr, #1
		mov	shiftl, shiftr
		subr	shiftl, #16

		mov	i, #3
		loc	ptrb, #resp
.loop2
		waitse1
		rdpin	pa, #PIN_CMD
		rev	pa
		mov	j, pa
		shr	pa, shiftr
		shl	pb, shiftl
		or	pb, pa
		movbyts	pb, #0b00_01
		wrword	pb, ptrb++
		mov	pb, j
		djnz	i, #.loop2

// DATA BLOCK RECEIVE
		dirl	#PIN_CLK
		dirl	#PIN_CMD
		wrpin	##P_SYNC_RX | (PIN_DAT2 - PIN_CMD & 7)<<28 | (PIN_CLK - PIN_CMD & 7)<<24, #PIN_CMD  // DAT2 pin
		wxpin	#31, #PIN_CMD    // DAT2 pin
		dirh	#PIN_DAT2 | 4<<6    // enable shifters for DAT0..DAT3 pins, and CLK pin
		wypin	##2000, #PIN_CLK    // kick it in the guts
.loop3
		waitse1    // trigger on PIN_CMD (DAT2 pin)
		rdpin	pr0, #PIN_DAT2    // DAT0 pin
		rdpin	pr1, #PIN_DAT3    // DAT1 pin
		rdpin	pr2, #PIN_CMD     // DAT2 pin
		rdpin	pr3, #PIN_LED     // DAT3 pin
		rev	pr2
		not	shiftr, pr2   wz
	if_z	jmp	#.loop3

		encod	shiftr
		mov	shiftl, shiftr
		subr	shiftl, #32

		rev	pr0
		rev	pr1
		rev	pr3
		mov	i, #32
		mov	ptrb, buf
.loop4
		waitse1
		rdpin	pr4, #PIN_DAT2    // DAT0 pin
		rdpin	pr5, #PIN_DAT3    // DAT1 pin
		rdpin	pr6, #PIN_CMD     // DAT2 pin
		rdpin	pr7, #PIN_LED     // DAT3 pin
		rev	pr4
		rev	pr5
		rev	pr6
		rev	pr7

		mov	pb, pr4
		shr	pb, shiftr
		shl	pr0, shiftl
		or	pr0, pb
		mov	pb, pr5
		shr	pb, shiftr
		shl	pr1, shiftl
		or	pr1, pb
		mov	pb, pr6
		shr	pb, shiftr
		shl	pr2, shiftl
		or	pr2, pb
		mov	pb, pr7
		shr	pb, shiftr
		shl	pr3, shiftl
		or	pr3, pb

		rolbyte	pb, pr3, #3
		rolbyte	pb, pr2, #3
		rolbyte	pb, pr1, #3
		rolbyte	pb, pr0, #3
		mergeb	pb
		movbyts	pb, #0b00_01_10_11
		wrlong	pb, ptrb++
		rolbyte	pb, pr3, #2
		rolbyte	pb, pr2, #2
		rolbyte	pb, pr1, #2
		rolbyte	pb, pr0, #2
		mergeb	pb
		movbyts	pb, #0b00_01_10_11
		wrlong	pb, ptrb++
		rolbyte	pb, pr3, #1
		rolbyte	pb, pr2, #1
		rolbyte	pb, pr1, #1
		rolbyte	pb, pr0, #1
		mergeb	pb
		movbyts	pb, #0b00_01_10_11
		wrlong	pb, ptrb++
		rolbyte	pb, pr3, #0
		rolbyte	pb, pr2, #0
		rolbyte	pb, pr1, #0
		rolbyte	pb, pr0, #0
		mergeb	pb
		movbyts	pb, #0b00_01_10_11
		wrlong	pb, ptrb++
		mov	pr0, pr4
		mov	pr1, pr5
		mov	pr2, pr6
		mov	pr3, pr7
		djnz	i, #.loop4
    }
    _pinf( SD_BASE_PIN | 7<<6 );    // All eight I/O pins of the ado-on header are used by the 4-bit SD board
    _wrpin( SD_BASE_PIN | 7<<6, 0 );
    _pinstart( PIN_CLK, P_PULSE | P_OE | P_INVERT_OUTPUT,    // SD clock gen smartpin
                CLK_DIV | CLK_DIV/2<<16, 0 );

    for( i = 0; i < 6; i++ )  {    // diag
        printf( " %02x", resp[i] );
    }
    printf( "  CRC = %02x\n", crc7(resp,5) );

    printf( " datablock:  shl=%d  shr=%d\n", shiftl, shiftr );
    for( i = 0; i <= 511/32; i++ )  {    // diag
        for( j = 0; j < 32; j++ )
            printf( " %02x", buf[i*32+j] );
        puts("");
    }

    return 1;
}

/*
              waitse2                           ' wait for first 16 bit pairs
              rdpin   rxd0,pinRxd0              ' MSW 16 bits, last received in bit 31
              rev     rxd0                      ' -> LSW, first bit received in bit 15
              zerox   rxd0,#15                  ' clear MSW
              encod   rxd1,rxd0                 ' find position of first '1'
              subr    rxd1,#30                  ' number of '0'bits = how many extra bits to shift
              wxpin   rxd1,pinRxdAll3           ' length = 16 data bit pairs + extra bits

' To synchronize the packet data we now need to shift more than 16 bit pairs in to get the full
' preamble + SFD. The bits of the last 16 bit pairs remain in the lower bits of the shift registers.
' The extra bits are at bit #15 down and push the leading '0' bits out at the right. The fresh
' bits are in the MSW and are now the synchronized last 32 bits of the preamble + SFD = $D5_555555.

              waitse2                           ' wait for rest of preamble received
              rdpin   rxd0,pinRxd0
              rdpin   rxd1,pinRxd1              ' shift registers -> MSWs
              cmp     rxd0,sfdEven wcz          ' check if preamble+SFD matches $D5_555555
              cmpx    rxd1,sfdOdd wz
              wxpin   #15,pinRxdAll3            ' set length back to 16 data bit pairs, again
        if_nz jmp     #rxAbort                  ' something is wrong
              mov     lenCnt,#maxBufLong-1      ' count longs to avoid overflow
*/


static uint32_t rxcmd( uint8_t * buf, size_t len )
{
    uint32_t  data;
    size_t  i;

    _pinstart( PIN_LED, P_REG_UP | P_OE | P_INVERT_OUTPUT |    // outputs to Activity LED, doubles up as rx start detection
                (PIN_CLK - PIN_LED & 7)<<28 |    // samrtA input select, count clock edges
                ((PIN_CMD - PIN_LED & 7) | 8)<<24,     // samrtB input select, inverted gating to count on lows
                0, 0 );
    _wrpin( PIN_CMD, P_SYNC_RX | (PIN_CLK - PIN_CMD & 7)<<24 );    // setup command responce smartpin mode
    _wxpin( PIN_CMD, 15 );    // 16-bit words, pre-rising clock sampling

    i = 64;    // max of 64 clocks for response starting timeout, SD spec 4.12.4
    do  {    // look for start-bit, one clock at a time
        _pinf( PIN_CMD );    // reset rx shifter
        _wypin( PIN_CLK, 1 );    // one SD clock
        _pinl( PIN_CMD );
        _waitx( CLK_DIV * 2 + 8 );    // clock smartpin cycle allowance + I/O lantency compensation
    } while( !_rdpin( PIN_LED ) && --i );

    if( _rdpin( PIN_LED ) )    // at least one bit counted
    {
        _wypin( PIN_CLK, len * 8 + 9 );    // response bit-clocks, SD spec 4.4 and 4.12.4
        i = 0;
        do  {
            if( _pinr( PIN_CMD ) )  {    // if data word is received
                data = _rev( _rdpin( PIN_CMD ) );    // collect rx data
                buf[i] = data >> 8;
                buf[i+1] = data;
                i += 2;
            }
        } while( i < len );
    }
    _pinf( PIN_CMD );
//    _wrpin( PIN_CMD, 0 );    // release CMD pin

    if( _rdpin( PIN_LED ) )  {
#ifdef _DIAG
        for( data = 0; data < len; data++ )  {    // diag
            printf( " %02x", buf[data] );
        }
#endif
        return 1;    // card has responded
    } else
        return 0;    // timed-out
}


static uint32_t sdcmd( uint32_t cmd, uint32_t arg )
{
    uint32_t  crc;

    if( cmd & 0x100 )  {    // check if it's an ACMD
        if( !sdcmd( 55, rca ) )    // ACMD's prefix
            return 0;
    }

    _wrpin( PIN_CMD, P_SYNC_TX | P_OE | (PIN_CLK - PIN_CMD & 7)<<24 );    // setup CMD output smartpin mode
    _wxpin( PIN_CMD, 31 );    // 32-bit shifter, continuous mode
    _wypin( PIN_CMD, _rev( (0x40 | cmd)<<24 | arg>>8 ) );    // first 32 bits into tx shifter, continuous mode
    _wypin( PIN_CLK, 6 * 8 );    // begin SD clocks, tx smartpin won't see clocks for about 8 sysclock ticks
    _pinl( PIN_CMD );    // start tx shifter, in continuous mode

    __asm {    // SD spec 4.5
		mov	pa, cmd
		or	pa, #0x40
		shl	pa, #24
		setq	pa
		mov	crc, #0
		crcnib	crc, #0x48    // CCITT polynomial is 1 + x3 + x7 (0x09 reversed for CRCNIB)
		crcnib	crc, #0x48
		setq	arg
		crcnib	crc, #0x48
		crcnib	crc, #0x48
		crcnib	crc, #0x48
		crcnib	crc, #0x48
		crcnib	crc, #0x48
		crcnib	crc, #0x48
		crcnib	crc, #0x48
		crcnib	crc, #0x48
		rev	crc
		shr	crc, #24
		or	crc, #1    // stop-bit at end of command packet
    }
    _wypin( PIN_CMD, _rev( arg<<24 | crc<<16 ) );    // final 16 bits into tx buffer
#ifdef _DIAG
    printf( "CMD%d   %02x %08x %02x  > ", cmd & 0xff, 0x40 | cmd & 0xff, arg, crc );    // diag
#endif

    while( !_pinr( PIN_CLK ) );    // wait for tx completion
    _pinf( PIN_CMD );
    _wrpin( PIN_CMD, 0 );    // release CMD pin, ready for response

    switch( cmd )  {
        case 2:
        case 9:
        case 10:
            crc = rxcmd( resp, 17 );
#ifdef _DIAG
            if( crc )
                crc = crc7( resp+1, 15 );
#endif
            break;
        case 3:
            crc = rxcmd( resp, 6 );
            if( crc )  {
#ifdef _DIAG
                crc = crc7( resp, 5 );
#endif
                rca = resp[1]<<24 | resp[2]<<16;
            }
            break;
        case 17:
            crc = rxblock( buff );
            break;
        default:
            crc = rxcmd( resp, 6 );
#ifdef _DIAG
            if( crc )
                crc = crc7( resp, 5 );
#endif
            break;
    }
    if( crc )  {
#ifdef _DIAG
        printf( "  CRC = %02x\n", crc );
#endif
        return( resp[1]<<24 | resp[2]<<16 | resp[3]<<8 | resp[4] );
    }
#ifdef _DIAG
    puts( " TIMED-OUT!" );
#endif
    return 0;
}


static int  sdinit( void )    // Targetted at controlling Roger Loh's 4-bit SD add-on board
{
    uint32_t  rc;

    _pinf( SD_BASE_PIN | 7<<6 );    // All eight I/O pins of the ado-on header are used by the 4-bit SD board
    _wrpin( SD_BASE_PIN | 7<<6, 0 );

    _waitms( 1 );
    if( !_pinr( PIN_CD ) )  {    // Card Detect will pull this down against the 22k ohm pull-up
        _pinh( PIN_CD );    // power off the SD card by overpowering the 1800 ohm Card Detect
        puts( " Card detected ... 40 ms power cycle of SD card ... " );
        _waitms( 40 );     // aproximately a 20 ms natural discharge curve for a plugged in card, SD spec 6.4.1
        _pinf( PIN_CD );     // and back on, released back to Card Detect function
        _waitms( 2 );
        puts( " DONE!" );

        _pinh( 56 );    // diag

        _pinstart( PIN_CLK, P_PULSE | P_OE | P_INVERT_OUTPUT,    // SD clock gen smartpin
                CLK_DIV | CLK_DIV/2<<16, 0 );
        _waitms( 1 );
        _wypin( PIN_CLK, 80 );    // SD spec 6.4.1
        _waitms( 1 );

        rca = 0;
        sdcmd( 0, 0 );    // abort to "idle" state - in case the power cycle reset wasn't successful

        rc = sdcmd( 8, 0x15a );    // SD spec 4.3.13
        if( rc == 0x15a )    // SD v2.0 valid voltage and check pattern
        {
            uint32_t  time, timer1 = _getms();
            uint32_t  timer2 = timer1;

            printf( "  R7 %08x - v2.0+ SD Card\n", rc );
            while(1)  {
                rc = sdcmd( 0x100|41, 1<<30 | 1<<20 );   // indicate HCS capable and 3v3 voltage
                if( rc>>31 )    // valid Ready bit, card has switched from "idle" to "ready" state
                    break;

                do  {
                    time = _getms();
                } while( time - timer2 < 40 );    // pause for 40 ms, SD spec 4.4
                timer2 += 40;
                if( time - timer1 > 1000 )    // SD spec 4.2.3
                    break;
            }
            printf( "  OCR %08x\n", rc );

            // There can be a CMD11 voltage switch here for controllers and cards that support it
            //   - leave this at 3V3 default, the Prop2 isn't suited to 1V8 signalling

            // Next two commands are explicitly ordered and required - singuluar operational path
            // If the sequence is not succesful then CMD3 will timeout and return a zero value
            sdcmd( 2, 0 );   // switch from "ready" to "ident" state, also next card status indicates a partial ACMD state
            if( rc = sdcmd( 3, 0 ) )    // switch from "ident" to "standby" state, also moves into "data transfer mode"
            {
                rca = rc & 0xffff_0000;
                printf( "  RCA %08x\n", rca );

                sdcmd( 7, rca );    // SELECT - switch from "standby" to "tran" state
                if( sdcmd( 0x100|6, 2 ) )  {    // Engage 4-bit parallel data mode
                    puts( " Card init complete" );
                    return  rca;
                }
            }
        }
    }
    puts( "Card not valid" );
    return  0;
}


static void  sdrelease( void )    // Targetted at controlling Roger Loh's 4-bit SD add-on board
{
    _pinf( SD_BASE_PIN | 7<<6 );    // All eight I/O pins of the ado-on header are used by the 4-bit SD board
    _wrpin( SD_BASE_PIN | 7<<6, 0 );
}


//===================================================================================

void  main( void )
{
    uint32_t  rc;

    printf( "   clkfreq = %d   clkmode = 0x%x\n", _clockfreq(), _clockmode() );

    rc = sdinit();  // returns a valid RCA with card in data transfer "standby" state, or zero for a fail

    if( rc )  {
    __asm {
		mov	rc, ##-1
		rdpin	rc, #PIN_CMD
    }
        printf( "Test RDPIN  %08x\n", rc );

        rc = sdcmd( 17, 0 );

        puts( "All finished  :)" );
    }

    _waitms( 500 );
    sdrelease();
    _clkset( 1, 20_000 );  // cool running
}
